import logging

import pytest
import yaml

from xmltotabular.sqlite_db import SqliteDB


@pytest.fixture
def empty_db():
    return SqliteDB(":memory:")


@pytest.fixture
def simple_config():
    config = """
    album:
      <entity>: album
      <fields>:
        name: name
        artist: artist
        released: released
        label: label
        genre: genre
    """
    return yaml.safe_load(config)


@pytest.fixture
def simple_data():
    return {
        "album": [
            {
                "id": "None_0",
                "name": "Five Leaves Left",
                "artist": "Nick Drake",
                "released": "1969",
                "label": "Island",
                "genre": "Folk",
            }
        ]
    }


@pytest.fixture
def debug_logger():
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())
    return logger



import logging

from xmltotabular import XmlCollectionToTabular


def test_simple_transform(simple_config):

    xml_path = "tests/test_xml/multiple_simple_docs.xml"

    collectionTransformer = XmlCollectionToTabular(
        xml_path,
        simple_config,
        ":memory:",
        "sqlite",
        validate=False,
        processes=1,
        continue_on_error=False,
        log_level=logging.DEBUG,
    )

    db = collectionTransformer.convert()

    selected = db.execute("SELECT * FROM album;").fetchall()

    assert selected == [
        ("Five Leaves Left", "Nick Drake", "1969", "Island", "Folk"),
        ("Bryter Layter", "Nick Drake", "1971", "Island", "Folk"),
        ("Pink Moon", "Nick Drake", "1972", "Island", "Folk"),
    ]



from xmltotabular import XmlDocToTabular


def test_simple_transform(simple_config, simple_data):
    xml = """\
<?xml version="1.0" encoding="UTF-8"?>
<album>
  <name>Five Leaves Left</name>
  <artist>Nick Drake</artist>
  <released>1969</released>
  <label>Island</label>
  <genre>Folk</genre>
</album>
    """

    docTransformer = XmlDocToTabular(simple_config)

    assert docTransformer.process_doc(xml) == simple_data



import itertools
import sqlite3
from collections import namedtuple

# > SQLITE_MAX_VARIABLE_NUMBER ... defaults to 999 for SQLite versions prior to 3.32.0
# > (2020-05-22) or 32766 for SQLite versions after 3.32.0.
#
# but many distributions ship SQLite compile with this value set much higher
#  (Debian-based Linux distros and Homebrew, for example, ship SQLite compiled with
#   SQLITE_MAX_VARIABLE_NUMBER set to 250,000).
#
# The default here is to assume the conservative value, but insert performance can be
#  increased significantly if a higher value can be used, so SqliteDB() accepts a user-
#  supplied value.  In modern versions of SQLite, the value set at compile-time can be
#  check with `echo "" | sqlite3 -cmd ".limits variable_number"`.
#
# See: https://www.sqlite.org/limits.html#max_variable_number
SQLITE_MAX_VARIABLE_NUMBER = 999

# See: https://www.sqlite.org/limits.html#max_column
SQLITE_MAX_COLUMN = 2000


COLUMN_TYPE_MAPPING = {
    float: "FLOAT",
    int: "INTEGER",
    bool: "INTEGER",
    str: "TEXT",
    bytes: "BLOB",
    # SQLite explicit types
    "TEXT": "TEXT",
    "INTEGER": "INTEGER",
    "FLOAT": "FLOAT",
    "BLOB": "BLOB",
    "text": "TEXT",
    "integer": "INTEGER",
    "float": "FLOAT",
    "blob": "BLOB",
}


class SqliteDB:
    def __init__(self, path, max_vars=None):
        self.path = path
        self.max_vars = max_vars or SQLITE_MAX_VARIABLE_NUMBER
        if path == ":memory:":
            self.conn = sqlite3.connect(":memory:")
        else:
            self.conn = sqlite3.connect(str(path), isolation_level=None)
        self.conn.execute("PRAGMA synchronous = OFF;")
        self.conn.execute("PRAGMA journal_mode = MEMORY;")
        self.conn.execute("PRAGMA locking_mode = EXCLUSIVE;")

    def __getitem__(self, table_name):
        return Table(self, table_name)

    def __repr__(self):
        return "<Database: {}>".format(self.path)

    def table_names(self):
        sql = "SELECT name FROM sqlite_master WHERE type = 'table';"
        return [r[0] for r in self.execute(sql).fetchall()]

    def execute(self, sql, parameters=None):
        if parameters is not None:
            return self.conn.execute(sql, parameters)
        else:
            return self.conn.execute(sql)

    def create_table_sql(self, name, columns, pk=None, column_order=None):
        validate_column_names(columns.keys())
        column_items = list(columns.items())
        if column_order is not None:
            column_items.sort(
                key=lambda p: column_order.index(p[0])
                if p[0] in column_order
                else SQLITE_MAX_COLUMN
            )

        column_defs = []
        if isinstance(pk, str):
            assert (
                pk in columns
            ), f"Specified primary-key ({pk}) is not in supplied columns!"
        for column_name, column_type in column_items:
            column_extras = []
            if column_name == pk:
                column_extras.append("PRIMARY KEY")
            column_defs.append(
                "   [{column_name}] {column_type}{column_extras}".format(
                    column_name=column_name,
                    column_type=COLUMN_TYPE_MAPPING[column_type],
                    column_extras=(" " + " ".join(column_extras))
                    if column_extras
                    else "",
                )
            )
        columns_sql = ",\n".join(column_defs)
        sql = f"CREATE TABLE [{name}] (\n{columns_sql}\n);"
        return sql

    def create_table(self, name, columns, pk=None, column_order=None):
        sql = self.create_table_sql(
            name=name,
            columns=columns,
            pk=pk,
            column_order=column_order,
        )
        self.execute(sql)
        return Table(self, name)


class Table:
    def __init__(self, db, name):
        self.db = db
        self.name = name

    @property
    def columns(self):
        rows = self.db.execute("PRAGMA table_info([{}])".format(self.name)).fetchall()
        return {row[1]: Column(*row) for row in rows}

    @property
    def schema(self):
        return self.db.execute(
            "SELECT sql FROM sqlite_master WHERE name = ?", (self.name,)
        ).fetchone()[0]

    def create(
        self,
        columns,
        pk=None,
        column_order=None,
    ):
        assert len(columns) <= min(self.db.max_vars, SQLITE_MAX_COLUMN), (
            f"Tables can have a maximum of {min(self.db.max_vars, SQLITE_MAX_COLUMN)} "
            "columns on this system."
        )

        with self.db.conn:
            self.db.create_table(
                self.name,
                columns,
                pk=pk,
                column_order=column_order,
            )
        return self

    def add_column(self, col_name, col_type=None):
        sql = "ALTER TABLE [{table}] ADD COLUMN [{col_name}] {col_type};".format(
            table=self.name, col_name=col_name, col_type=COLUMN_TYPE_MAPPING[col_type]
        )
        self.db.execute(sql)
        return self

    def generate_insert_batches(self, records):
        def batches(records, max_batch_size):
            """Yield successive batches of `records` of size `max_batch_size`."""
            for i in range(0, len(records), max_batch_size):
                yield records[i : i + max_batch_size]

        num_columns = len(self.columns)

        max_batch_size = self.db.max_vars // num_columns
        columns = ", ".join(f"[{c}]" for c in self.columns)
        placeholders = ", ".join("?" * num_columns)

        for batch in batches(records, max_batch_size):
            params = [
                [record.get(key, None) for key in self.columns] for record in batch
            ]
            rows = ", ".join(f"({placeholders})" for record in batch)
            sql = f"INSERT INTO [{self.name}] ({columns}) VALUES {rows};"
            yield (sql, list(itertools.chain(*params)))

    def insert_all(self, records):
        for sql, params in self.generate_insert_batches(records):
            self.db.execute(sql, params)
        self.db.conn.commit()


Column = namedtuple(
    "Column", ("cid", "name", "type", "notnull", "default_value", "is_pk")
)


def validate_column_names(column_names):
    # Columns may not contain '[' or ']' (https://bugs.python.org/issue39652)
    for column_name in column_names:
        assert (
            "[" not in column_name and "]" not in column_name
        ), "'[' and ']' cannot be used in column names"



import pytest
import re
import tempfile
from pathlib import Path

from xmltotabular.sqlite_db import (
    SqliteDB,
    SQLITE_MAX_VARIABLE_NUMBER,
    SQLITE_MAX_COLUMN,
)
from xmltotabular.utils import get_fieldnames_from_config


def normalize_schema(schema):
    """Normalize whitespace in SQL statements returned by SqliteDB[table].schema
    for convenient comparison."""
    return re.sub(r"\s(?=[,)])|(?<=\()\s", "", " ".join(schema.split()))


def test_sqlitedb_repr(empty_db):
    assert repr(empty_db) == "<Database: :memory:>"


def test_simple_table_creation(empty_db, simple_config):
    db = empty_db

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        db[tablename].create({fieldname: str for fieldname in fieldnames})

    expected_schema = """
    CREATE TABLE [album] (
        [name] TEXT,
        [artist] TEXT,
        [released] TEXT,
        [label] TEXT,
        [genre] TEXT
    )
    """

    assert "album" in db.table_names()
    assert normalize_schema(db["album"].schema) == normalize_schema(expected_schema)


def test_table_creation_with_fields_reversed(empty_db, simple_config):
    db = empty_db

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        params = {"column_order": list(reversed(fieldnames))}
        db[tablename].create({fieldname: str for fieldname in fieldnames}, **params)

    expected_schema = """
    CREATE TABLE [album] (
        [genre] TEXT,
        [label] TEXT,
        [released] TEXT,
        [artist] TEXT,
        [name] TEXT
    )
    """

    assert "album" in db.table_names()
    assert normalize_schema(db["album"].schema) == normalize_schema(expected_schema)


def test_table_creation_with_pk(empty_db, simple_config):
    db = empty_db

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        params = {"pk": "name"}
        db[tablename].create({fieldname: str for fieldname in fieldnames}, **params)

    expected_schema = """
    CREATE TABLE [album] (
        [name] TEXT PRIMARY KEY,
        [artist] TEXT,
        [released] TEXT,
        [label] TEXT,
        [genre] TEXT
    )
    """

    assert "album" in db.table_names()
    assert normalize_schema(db["album"].schema) == normalize_schema(expected_schema)


def test_table_creation_additional_fields(empty_db, simple_config):
    db = empty_db

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        db[tablename].create({fieldname: str for fieldname in fieldnames})

    additional_fields = {"album": ["description", "notes"]}
    for tablename, fieldnames in additional_fields.items():
        if tablename in db.table_names():
            for fieldname in fieldnames:
                if fieldname not in db[tablename].columns:
                    db[tablename].add_column(fieldname, str)

    expected_schema = """
    CREATE TABLE [album] (
        [name] TEXT,
        [artist] TEXT,
        [released] TEXT,
        [label] TEXT,
        [genre] TEXT,
        [description] TEXT,
        [notes] TEXT
    )
    """

    assert "album" in db.table_names()
    assert normalize_schema(db["album"].schema) == normalize_schema(expected_schema)


def test_simple_data_insertion(empty_db, simple_config, simple_data):
    db = empty_db

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        db[tablename].create({fieldname: str for fieldname in fieldnames})

    for tablename, rows in simple_data.items():
        db[tablename].insert_all(rows)

    selected = db.execute("SELECT * FROM album;").fetchall()

    assert selected == [("Five Leaves Left", "Nick Drake", "1969", "Island", "Folk")]


def test_multiple_rows_insertion(empty_db):
    db = empty_db
    tablename = "test-table"
    num_columns = 10
    num_rows = 100

    db[tablename].create({f"column_{col:02}": str for col in range(num_columns)})

    db[tablename].insert_all(
        [
            {f"column_{col:02}": f"data_{row:03}_{col:02}" for col in range(num_columns)}
            for row in range(num_rows)
        ]
    )

    selected = db.execute(f"SELECT * FROM '{tablename}';").fetchall()

    assert selected == [
        tuple(f"data_{row:03}_{col:02}" for col in range(num_columns))
        for row in range(num_rows)
    ]


def test_writing_to_filesystem(simple_config, simple_data):
    dirpath = tempfile.mkdtemp()
    db_path = Path(dirpath) / "test_db.sqlite"

    db = SqliteDB(db_path)

    for tablename, fieldnames in get_fieldnames_from_config(simple_config).items():
        db[tablename].create({fieldname: str for fieldname in fieldnames})

    for tablename, rows in simple_data.items():
        db[tablename].insert_all(rows)

    selected = db.execute("SELECT * FROM album;").fetchall()

    assert selected == [("Five Leaves Left", "Nick Drake", "1969", "Island", "Folk")]


@pytest.mark.parametrize(
    "num_columns,should_error",
    (
        (100, False),
        (min(SQLITE_MAX_VARIABLE_NUMBER, SQLITE_MAX_COLUMN), False),
        (min(SQLITE_MAX_VARIABLE_NUMBER, SQLITE_MAX_COLUMN) + 1, True),
    ),
)
def test_error_if_too_many_columns(empty_db, num_columns, should_error):
    columns = {f"c{i}": str for i in range(num_columns)}
    if should_error:
        with pytest.raises(AssertionError):
            empty_db["too-many-columns"].create(columns)
    else:
        empty_db["too-many-columns"].create(columns)


@pytest.mark.parametrize(
    "num_columns,max_vars,should_error",
    (
        (100, 250_000, False),
        (SQLITE_MAX_COLUMN, 250_000, False),
        (SQLITE_MAX_COLUMN + 1, 250_000, True),
    ),
)
def test_error_if_too_many_columns_with_custom_max_vars(
    num_columns, max_vars, should_error
):
    db = SqliteDB(":memory:", max_vars=max_vars)
    columns = {f"c{i}": str for i in range(num_columns)}
    if should_error:
        with pytest.raises(AssertionError):
            db["too-many-columns"].create(columns)
    else:
        db["too-many-columns"].create(columns)


@pytest.mark.parametrize(
    "num_rows",
    (
        # Simplest case
        1,
        # Default SQLITE_MAX_VARIABLE_NUMBER for SQLite versions < 3.32.0 (2020-05-22)
        999,
        999 + 1,
        # Default SQLITE_MAX_VARIABLE_NUMBER for SQLite versions >= 3.32.0
        32766 + 1,
        # Default SQLITE_MAX_VARIABLE_NUMBER for SQLite distributed
        #  with Debian, Ubuntu, Homebrew etc.
        250000 + 1,
    ),
)
def test_error_if_too_many_vars(empty_db, num_rows):
    empty_db["too-many-vars"].create({"c": str})
    rows = [{"c": i} for i in range(num_rows)]
    empty_db["too-many-vars"].insert_all(rows)



import logging
import re
from collections import defaultdict
from functools import partial
from io import BytesIO
from pprint import pformat

from lxml import etree

from .utils import (
    DTDResolver,
    colored,
    test_doctype,
    WrongDoctypeException,
    NoDoctypeException,
)


class XmlDocToTabular:
    def __init__(
        self,
        config,
        logger=None,
        dtd_path=None,
        preprocess_doc=None,
        validate=False,
        continue_on_error=False,
        log_level=None,
    ):
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.StreamHandler())

        if log_level:
            self.logger.setLevel(log_level)

        self.config = config
        self.dtd_path = dtd_path
        self.preprocess_doc = preprocess_doc
        self.validate = validate
        self.continue_on_error = continue_on_error
        self.tables = defaultdict(list)
        # lambdas can't be pickled (without dill, at least)
        self.table_pk_idx = defaultdict(partial(defaultdict, int))

    @staticmethod
    def get_text(xpath_result):
        if isinstance(xpath_result, str):
            return re.sub(r"\s+", " ", xpath_result).strip()
        return re.sub(
            r"\s+", " ", etree.tostring(xpath_result, method="text", encoding="unicode")
        ).strip()

    def get_pk(self, tree, config):
        def get_pk_component(expression):
            elems = tree.xpath(f"./{expression}")
            assert (
                len(elems) == 1
            ), f"{len(elems)} elements found for <primary_key> component {expression}"
            return self.get_text(elems[0])

        if "<primary_key>" in config and isinstance(config["<primary_key>"], str):
            return get_pk_component(config["<primary_key>"])

        if "<primary_key>" in config and isinstance(config["<primary_key>"], list):
            return "-".join(
                get_pk_component(expression) for expression in config["<primary_key>"]
            )

        return None

    def add_string(self, path, elems, record, fieldname):
        try:
            assert len(elems) == 1
        except AssertionError as exc:
            exc.msg = (
                f"Multiple elements found for {path}! "
                + "Should your config file include a joiner, or new entity "
                + "definition?"
                + "\n\n- "
                + "\n- ".join(self.get_text(el) for el in elems)
            )
            raise

        # we've only one elem, and it's a simple mapping to a fieldname
        record[fieldname] = self.get_text(elems[0])

    def process_doc_from_pool(self, payload):
        """Unpack a tuple returned by yield_xml_doc().

        This is necessary because multiprocessing.Pool.imap will only pass a single
        argument to child processes. Pool.starmap would be great, but there's no version
        that works with iterables, which is really needed here. Pathos' multiprocess can
        handle this (and is required by this library when using python 3.6 anyway) --
        there may be value in using this anyway.
        """
        return self.process_doc(**payload)

    def process_doc(self, doc, filename=None, linenum=None):

        if "xml_root" in self.config:
            try:
                test_doctype(doc, self.config["xml_root"])

            except WrongDoctypeException as exc:
                self.logger.debug(
                    "%s\n%s",
                    colored(
                        "Unexpected XML document"
                        + (f" ending at line {linenum}" if linenum else "")
                        + (f" in file {filename}" if filename else "")
                        + ": ",
                        "yellow",
                    ),
                    str(exc),
                )
                return self.tables

            except NoDoctypeException:
                self.logger.debug(
                    "%s\n%s",
                    colored(
                        "Document"
                        + (f" ending at line {linenum}" if linenum else "")
                        + (f" in file {filename}" if filename else "")
                        + " has no DOCTYPE:",
                        "yellow",
                    ),
                    doc,
                )
                return self.tables

        try:
            tree = self.parse_tree(doc)
            for path, config in self.config.items():
                self.process_path(tree, path, config, filename, {})

        except LookupError as exc:
            self.logger.warning(exc.args[0])
            if not self.continue_on_error:
                raise SystemExit()

        except etree.XMLSyntaxError as exc:
            self.logger.warning(
                colored(
                    "Unable to parse XML document"
                    + (f" ending at line {linenum}" if linenum else "")
                    + (f" in file {filename}" if filename else "")
                    + " (enable debug logging to dump doc to console):",
                    "red",
                )
                + colored(f"\n    {exc.msg}", "yellow")
            )
            self.logger.debug(doc)

            if not self.continue_on_error:
                raise SystemExit()

        except AssertionError as exc:
            pk = self.get_pk(self.parse_tree(doc), next(iter(self.config.values())))
            self.logger.warning(
                colored(
                    "Unable to parse document"
                    + (f" with ID {pk}" if pk else "")
                    + (f" ending at line {linenum}" if linenum else "")
                    + (f" in file {filename}" if filename else "")
                    + " -- record has not been parsed"
                    + " (enable debug logging to dump doc to console):",
                    "red",
                )
                + colored(f"\n    {exc.msg}", "yellow")
            )
            self.logger.debug(doc)

            if not self.continue_on_error:
                raise SystemExit()

        return self.tables

    def parse_tree(self, doc):
        if self.preprocess_doc:
            doc = self.preprocess_doc(doc)

        parser_args = {
            "load_dtd": True,
            "resolve_entities": True,
            "ns_clean": True,
            "huge_tree": True,
            "collect_ids": False,
        }

        if self.validate:
            parser_args["dtd_validation"] = True

        parser = etree.XMLParser(**parser_args)
        if self.dtd_path:
            parser.resolvers.add(DTDResolver(self.dtd_path))
        return etree.parse(BytesIO(doc.encode("utf8")), parser)

    def process_path(
        self, tree, path, config, filename, record, parent_entity=None, parent_pk=None
    ):
        try:
            elems = [tree.getroot()]
        except AttributeError:
            elems = tree.xpath("./" + path)

        self.process_field(
            elems, tree, path, config, filename, record, parent_entity, parent_pk
        )

    def process_field(
        self,
        elems,
        tree,
        path,
        config,
        filename,
        record,
        parent_entity=None,
        parent_pk=None,
    ):

        if isinstance(config, str):
            if elems:
                self.add_string(path, elems, record, config)
            return

        if "<entity>" in config:
            # config is a new entity definition (i.e. a new record on a new table/file)
            self.process_new_entity(
                tree, elems, config, filename, parent_entity, parent_pk
            )
            return

        if "<fieldname>" in config:
            # config is extra configuration for a field on this table/file
            if "<joiner>" in config:
                if elems:
                    record[config["<fieldname>"]] = config["<joiner>"].join(
                        [self.get_text(elem) for elem in elems]
                    )
                return

            if "<enum_map>" in config:
                if elems:
                    record[config["<fieldname>"]] = config["<enum_map>"].get(
                        self.get_text(elems[0])
                    )
                return

            if "<enum_type>" in config:
                if elems:
                    record[config["<fieldname>"]] = config["<enum_type>"]
                return

            # just a mapping to a fieldname string
            if len(config) == 1:
                self.add_string(path, elems, record, config["<fieldname>"])
                return

        # We may have multiple configurations for this key (XPath expression)
        if isinstance(config, list):
            for subconfig in config:
                self.process_field(
                    elems,
                    tree,
                    path,
                    subconfig,
                    filename,
                    record,
                    parent_entity,
                    parent_pk,
                )
            return

        raise LookupError(
            f'Invalid configuration for key "{parent_entity}":'
            + "\n "
            + "\n ".join(pformat(config).split("\n"))
        )

    def process_new_entity(
        self, tree, elems, config, filename, parent_entity=None, parent_pk=None
    ):
        """Process a subtree of the xml as a new entity type, creating a new record in a
        new output table/file.
        """
        entity = config["<entity>"]
        for elem in elems:
            record = {}

            pk = self.get_pk(tree, config)
            if pk:
                record["id"] = pk
            else:
                record["id"] = f"{parent_pk}_{self.table_pk_idx[entity][parent_pk]}"
                self.table_pk_idx[entity][parent_pk] += 1

            if parent_pk:
                record[f"{parent_entity}_id"] = parent_pk
            if "<filename_field>" in config:
                record[config["<filename_field>"]] = filename
            for subpath, subconfig in config["<fields>"].items():
                self.process_path(elem, subpath, subconfig, filename, record, entity, pk)

            self.tables[entity].append(record)



import csv
import logging
import sys
from collections import defaultdict
from pathlib import Path

import yaml

from .sqlite_db import SqliteDB
from .utils import (
    expand_paths,
    colored,
    Pool,
    cpu_count,
    yield_xml_doc,
    get_fieldnames_from_config,
)
from .xmldoctotabular import XmlDocToTabular


class XmlCollectionToTabular:
    def __init__(
        self,
        xml_input,
        config,
        output_path,
        output_type="sqlite",
        dtd_path=None,
        preprocess_doc=None,
        log_level=logging.INFO,
        recurse=True,
        validate=False,
        check_doctype=False,
        processes=None,
        continue_on_error=False,
        sqlite_max_vars=None,
    ):

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(log_level)
        self.logger.addHandler(logging.StreamHandler(sys.stdout))

        self.xml_files = []
        if isinstance(xml_input, str):
            xml_input = [xml_input]
        for input_path in xml_input:
            for path in expand_paths(input_path):
                if path.is_file():
                    self.xml_files.append(path)
                elif path.is_dir():
                    self.xml_files.extend(
                        path.glob(f'{"**/" if recurse else ""}*.[xX][mM][lL]')
                    )
                else:
                    self.logger.fatal("specified input is invalid")
                    exit(1)

        self.config = config
        if isinstance(config, str) and Path(config).is_file():
            self.config = yaml.safe_load(open(config))

        self.output_type = output_type
        self.output_path = output_path

        if self.output_type == "sqlite" and self.output_path == ":memory:":
            self.init_sqlite_db(self.output_path)

        elif self.output_type == "sqlite":
            self.output_path = Path(self.output_path).resolve()

            if self.output_path.is_dir():
                self.output_path = (self.output_path / "db.sqlite").resolve()

            if self.output_path.suffix != "sqlite":
                self.output_path = self.output_path.with_suffix(".sqlite")

            if self.output_path.exists():
                self.logger.warning(
                    colored(
                        "Sqlite database %s exists; records will be appended.",
                        "yellow",
                    ),
                    self.output_path,
                )
            else:
                self.output_path.parent.mkdir(parents=True, exist_ok=True)

            self.init_sqlite_db(self.output_path)

        else:
            self.output_path = Path(self.output_path).resolve()
            self.output_path.mkdir(parents=True, exist_ok=True)

        self.dtd_path = dtd_path
        self.preprocess_doc = preprocess_doc
        self.validate = validate
        self.processes = processes
        self.continue_on_error = continue_on_error

        self.fieldnames = get_fieldnames_from_config(self.config)
        if check_doctype:
            self.set_root_config()

    def set_root_config(self):
        if "xml_root" not in self.config:
            self.config["xml_root"] = next(iter(self.config.keys()))
            self.logger.warning(
                colored(
                    "<xml_root> not explicitly set in config -- assuming <%s/>",
                    "yellow",
                ),
                self.config["xml_root"],
            )

    def init_sqlite_db(self, output_path):
        self.db = SqliteDB(output_path)

        for tablename, fieldnames in get_fieldnames_from_config(self.config).items():
            if tablename in self.db.table_names():
                for fieldname in fieldnames:
                    if fieldname not in self.db[tablename].columns:
                        self.db[tablename].add_column(fieldname, str)
                continue
            params = {"column_order": fieldnames}
            if "id" in fieldnames:
                params["pk"] = "id"
            self.db[tablename].create(
                {fieldname: str for fieldname in fieldnames}, **params
            )

    def convert(self):
        if not self.xml_files:
            self.logger.warning(colored("No input files to process!", "red"))

        docParser = XmlDocToTabular(
            logger=self.logger,
            config=self.config,
            dtd_path=self.dtd_path,
            preprocess_doc=self.preprocess_doc,
            validate=self.validate,
            continue_on_error=self.continue_on_error,
        )

        for input_file in self.xml_files:

            self.logger.warning(
                colored("Processing %s...", "green"), input_file.resolve()
            )

            processes = self.processes or cpu_count() - 1 or 1
            # chunk sizes greater than 1 result in duplicate returns because the results
            #  are pooled on the XmlDocToTabular instance
            chunksize = 1

            pool = Pool(processes=processes)

            all_tables = defaultdict(list)
            for i, tables in enumerate(
                pool.imap(
                    docParser.process_doc_from_pool,
                    yield_xml_doc(input_file),
                    chunksize,
                )
            ):

                if i % 100 == 0:
                    self.logger.info(colored("Processing document %d...", "cyan"), i + 1)
                for key, value in tables.items():
                    all_tables[key].extend(value)

            pool.close()
            pool.join()

            if all_tables:
                self.logger.info(colored("...%d records processed!", "green"), i + 1)
                self.write_tables(all_tables)
            else:
                self.logger.warning(
                    colored("No records found! (config file error?)", "red")
                )

        if self.output_type == "sqlite" and self.output_path == ":memory:":
            return self.db

    def write_tables(self, tables):
        if self.output_type == "csv":
            self.write_csv_files(tables)

        if self.output_type == "sqlite":
            self.write_sqlitedb(tables)

    def write_csv_files(self, tables):

        self.logger.info(
            "%s",
            colored(f"Writing csv files to {self.output_path.resolve()} ...", "green"),
        )
        for tablename, rows in tables.items():
            output_file = self.output_path / f"{tablename}.csv"

            if output_file.exists():
                self.logger.debug(
                    "%s",
                    colored(
                        f"CSV file {output_file} exists; records will be appended.",
                        "yellow",
                    ),
                )

                with output_file.open("a") as _fh:
                    writer = csv.DictWriter(_fh, fieldnames=self.fieldnames[tablename])
                    writer.writerows(rows)

            else:
                with output_file.open("w") as _fh:
                    writer = csv.DictWriter(_fh, fieldnames=self.fieldnames[tablename])
                    writer.writeheader()
                    writer.writerows(rows)

    def write_sqlitedb(self, tables):
        self.logger.info(colored("Writing records to %s ...", "green"), self.output_path)
        self.db.conn.execute("begin exclusive;")
        for tablename, rows in tables.items():
            self.logger.info(
                colored("Writing %d records to `%s`...", "magenta"),
                len(rows),
                tablename,
            )
            self.db[tablename].insert_all(rows)



import sys
from collections import defaultdict
from pathlib import Path
from pprint import pformat

from lxml import etree

assert sys.version_info >= (3, 6), "Error: Python 3.6 or newer is required."

if sys.version_info < (3, 7):
    try:
        from multiprocess import Pool, cpu_count
    except ImportError:
        sys.exit(
            "Error: If running with Python < 3.7, the multiprocess library is required "
            "(e.g. pip install multiprocess)."
        )
else:
    from multiprocessing import Pool, cpu_count  # noqa: F401


try:
    assert sys.stdout.isatty()
    from termcolor import colored
except (AssertionError, ImportError):

    def colored(text, *args, **kwargs):
        """Dummy function to pass text through without escape codes if stdout is not a
        TTY or termcolor is not available."""
        return text


def expand_paths(path_expr):
    """Given a path expression, return a list of paths."""
    path = Path(path_expr).expanduser()
    if path.is_file():
        return [path]
    return Path(path.root).glob(
        str(Path("").joinpath(*path.parts[1:] if path.is_absolute() else path.parts))
    )


class DTDResolver(etree.Resolver):
    """A DTDResolver class which resolves DTDs relative to a supplied base path."""

    def __init__(self, dtd_path):
        self.dtd_path = Path(dtd_path)

    def resolve(self, system_url, _public_id, context):
        if system_url.startswith(str(self.dtd_path)):
            return self.resolve_filename(system_url, context)
        else:
            return self.resolve_filename(
                str((self.dtd_path / system_url).resolve()), context
            )


class WrongDoctypeException(Exception):
    """Used to indicate that an XML doc has the wrong DOCTYPE."""


class NoDoctypeException(Exception):
    """Used to indicate that an XML doc does not specify a DOCTYPE."""


def test_doctype(doc, expected_doctype):
    """Test an XML document (passed as a string) for a DOCTYPE declaration that matches
    `expected_doctype`."""
    for line in doc.split("\n"):
        if line.startswith(f"<!DOCTYPE {expected_doctype}"):
            return True
        elif line.startswith("<!DOCTYPE "):
            raise WrongDoctypeException(line)

    raise NoDoctypeException()


def yield_xml_doc(filepath):
    """Given a path to a file containing one or more XML documents, for each document
    yield a dictionary containing a document, the filename, and the ending line number
    at which the document is found."""
    filename = filepath.resolve().name
    xml_doc = []

    with open(filepath, "r", errors="replace") as _fh:
        for i, line in enumerate(_fh):
            if xml_doc and line.startswith("<?xml "):
                yield {
                    "filename": filename,
                    "linenum": i - len(xml_doc),
                    "doc": "".join(xml_doc),
                }
                xml_doc = []

            # handle the case where documents have been concatenated without
            #  adequate interpolation of new-lines
            elif xml_doc and "<?xml " in line:
                xml_doc.append(line[: line.find("<?xml ")])
                line = line[line.find("<?xml ") :]
                yield {
                    "filename": filename,
                    "linenum": i - len(xml_doc),
                    "doc": "".join(xml_doc),
                }
                xml_doc = []

            xml_doc.append(line)

        yield {
            "filename": filename,
            "linenum": i - len(xml_doc),
            "doc": "".join(xml_doc),
        }


def get_fieldnames_from_config(full_config):
    """Parse a config object and return a dictionary where keys are table names and
    values are lists of field names."""

    # On python >=3.7, dictionaries maintain key order, so fields are guaranteed to
    #  be returned in the order in which they appear in the config file.  To
    #  guarantee this on versions of python <3.7 (insofar as it matters),
    #  collections.OrderedDict would have to be used here.

    fieldnames = defaultdict(list)

    def add_fieldnames(config, _fieldnames, parent_entity=None):
        if isinstance(config, str):
            if ":" in config:
                _fieldnames.append(config.split(":")[0])
                return
            _fieldnames.append(config)
            return

        if "<fieldname>" in config:
            _fieldnames.append(config["<fieldname>"])
            return

        if "<entity>" in config:
            entity = config["<entity>"]
            _fieldnames = []
            if "<primary_key>" in config or parent_entity:
                _fieldnames.append("id")
            if parent_entity:
                _fieldnames.append(f"{parent_entity}_id")
            if "<filename_field>" in config:
                _fieldnames.append(config["<filename_field>"])
            for subconfig in config["<fields>"].values():
                add_fieldnames(subconfig, _fieldnames, entity)
            # different keys (XPath expressions) may be appending rows to the same
            #  table(s), so we're appending to lists of fieldnames here.
            fieldnames[entity] = list(
                dict.fromkeys(fieldnames[entity] + _fieldnames).keys()
            )
            return

        # We may have multiple configurations for this key (XPath expression)
        if isinstance(config, list):
            for subconfig in config:
                add_fieldnames(subconfig, _fieldnames, parent_entity)
            return

        raise LookupError(
            "Invalid configuration:" + "\n " + "\n ".join(pformat(config).split("\n"))
        )

    for key, config in full_config.items():
        if key.startswith("<"):
            # skip keyword instructions
            continue
        add_fieldnames(config, [])

    return fieldnames
